module FastIO
    (
     Handle,

     openFile,
     closeFile,
     isEOF,
     fileError,

     inspectChar,

     unsafe_fscanfInt,
     unsafe_fscanfChar,
     unsafe_fscanfFloat,

     unsafe_freadInt,
     unsafe_freadChar,
     unsafe_freadFloat,

     fprintfInt,
     fprintfChar,
     fprintfFloat,
     fprintfString,

     fwriteInt,
     fwriteChar,
     fwriteFloat,

     fscanfInt,
     fscanfChar,
     fscanfFloat
    )where

import Foreign
import Foreign.C.Types
import qualified System.IO as IO
import Data.IORef

import Data.Char

newtype Handle = H (Ptr ())

-- exported primitive function definitions

openFile :: FilePath -> IO.IOMode -> IO Handle
openFile fp mode = do
  fp' <- cstring fp
  md' <- mkMode mode
  h   <- c__openFile fp' md'
  return (H h)

closeFile :: Handle -> IO ()
closeFile (H h) = c__closeFile h

isEOF :: Handle -> IO Bool
isEOF (H h) = do v <- c__isEOF h; return (v /= 0);

fileError :: Handle -> IO (Maybe Int)
fileError (H h) = do
  v <- c__fileError h
  if v == 0 
    then return Nothing
    else return (Just $ fromIntegral v)

unsafe_fscanfInt :: Handle -> IO Int
unsafe_fscanfInt (H h) = c__unsafe_fscanfInt h >>= return . fromIntegral

unsafe_fscanfChar :: Handle -> IO Char
unsafe_fscanfChar (H h) = do i <- c__unsafe_fscanfChar h; return (chr $ fromIntegral i)

unsafe_fscanfFloat :: Handle -> IO Float
unsafe_fscanfFloat (H h) = c__unsafe_fscanfFloat h >>= unsafeCast

unsafe_freadInt :: Handle -> IO Int
unsafe_freadInt (H h) = c__unsafe_freadInt h >>= return . fromIntegral

unsafe_freadChar :: Handle -> IO Char
unsafe_freadChar (H h) = do i <- c__unsafe_freadChar h; return (chr $ fromIntegral i)

unsafe_freadFloat :: Handle -> IO Float
unsafe_freadFloat (H h) = c__unsafe_freadFloat h >>= unsafeCast

fprintfInt :: Handle -> Int -> IO ()
fprintfInt (H h) i = c__fprintfInt h (fromIntegral i)

fprintfChar :: Handle -> Char -> IO ()
fprintfChar (H h) c = unsafeCast ((fromIntegral $ ord c) :: Int8) >>= c__fprintfChar h

fprintfFloat :: Handle -> Float -> IO ()
fprintfFloat (H h) f = unsafeCast f >>= c__fprintfFloat h

fprintfString :: Handle -> String -> IO ()
fprintfString (H h) s = cstring s >>= c__fprintfString h

inspectChar :: Handle -> IO Char
inspectChar (H h) = do c <- c__inspectChar h; return (chr $ fromIntegral c)

fwriteInt :: Handle -> Int -> IO ()
fwriteInt (H h) i = c__fwriteInt h (fromIntegral i)

fwriteChar :: Handle -> Char -> IO ()
fwriteChar (H h) c = unsafeCast ((fromIntegral $ ord c) :: Int8) >>= c__fwriteChar h

fwriteFloat :: Handle -> Float -> IO ()
fwriteFloat (H h) f = unsafeCast f >>= c__fwriteFloat h

-- exported non-primitive function definitions

fscanfInt :: Handle -> IO Int
fscanfInt h = do
  v <- unsafe_fscanfInt h
  e <- fileError h
  case e of
    Nothing -> return v
    Just ec -> ioError $ userError ("fscanfInt produced error: " ++ show ec)

fscanfChar :: Handle -> IO Char
fscanfChar h = do
  v <- unsafe_fscanfChar h
  e <- fileError h
  case e of
    Nothing -> return v
    Just ec -> ioError $ userError ("fscanfChar produced error: " ++ show ec)

fscanfFloat :: Handle -> IO Float
fscanfFloat h = do
  v <- unsafe_fscanfFloat h
  e <- fileError h
  case e of
    Nothing -> return v
    Just ec -> ioError $ userError ("fscanfFloat produced error: " ++ show ec)

-- c interface through ffi

foreign import ccall "openFile"  c__openFile        :: Ptr CChar -> Ptr CChar -> IO (Ptr ())
foreign import ccall "closeFile" c__closeFile       :: Ptr () -> IO ()
foreign import ccall "isEOF"     c__isEOF           :: Ptr () -> IO CInt

foreign import ccall "inspectChar" c__inspectChar           :: Ptr () -> IO CChar

foreign import ccall "unsafe_fscanfInt"    c__unsafe_fscanfInt    :: Ptr () -> IO CInt
foreign import ccall "unsafe_fscanfChar"   c__unsafe_fscanfChar   :: Ptr () -> IO CChar
foreign import ccall "unsafe_fscanfFloat"  c__unsafe_fscanfFloat  :: Ptr () -> IO CFloat

foreign import ccall "unsafe_freadInt"    c__unsafe_freadInt    :: Ptr () -> IO CInt
foreign import ccall "unsafe_freadChar"   c__unsafe_freadChar   :: Ptr () -> IO CChar
foreign import ccall "unsafe_freadFloat"  c__unsafe_freadFloat  :: Ptr () -> IO CFloat

foreign import ccall "fprintfInt"    c__fprintfInt    :: Ptr () -> CInt   -> IO ()
foreign import ccall "fprintfChar"   c__fprintfChar   :: Ptr () -> CChar  -> IO ()
foreign import ccall "fprintfFloat"  c__fprintfFloat  :: Ptr () -> CFloat -> IO ()
foreign import ccall "fprintfString" c__fprintfString :: Ptr () -> Ptr CChar -> IO ()

foreign import ccall "fwriteInt"    c__fwriteInt    :: Ptr () -> CInt   -> IO ()
foreign import ccall "fwriteChar"   c__fwriteChar   :: Ptr () -> CChar  -> IO ()
foreign import ccall "fwriteFloat"  c__fwriteFloat  :: Ptr () -> CFloat -> IO ()

foreign import ccall "fileError" c__fileError       :: Ptr () -> IO CInt

-- undefined interfact through nothing
{-
c__openFile :: Ptr CChar -> Ptr CChar -> IO (Ptr ())
c__openFile = undefined

c__closeFile :: Ptr () -> IO ()
c__closeFile = undefined

c__unsafe_fscanfInt :: Ptr () -> IO CInt
c__unsafe_fscanfInt = undefined
-}

------- helper functions -------

mkMode IO.ReadMode      = cstring "r"
mkMode IO.WriteMode     = cstring "w"
mkMode IO.AppendMode    = cstring "a"
mkMode IO.ReadWriteMode = cstring "rw"

cstring s = do
  arr <- mallocArray0 (length s)
  (cchars :: [CChar]) <- mapM unsafeCast s
  pokeArray0 (0 :: CChar) arr cchars
  return arr


unsafeCastHelper :: IORef a
unsafeCastHelper = unsafePerformIO (newIORef undefined)

unsafeCast :: a -> IO b
unsafeCast x = 
  writeIORef unsafeCastHelper x >>
  readIORef unsafeCastHelper
